@@ -212,20 +212,6 @@ static thread_return_t THREAD_CALL resolver_thread_entry(void *arg) {
 	return (thread_return_t)0;
 }
 
-void agent_add_ice_tcp_local_candidate(juice_agent_t *agent, addr_record_t *record) {
-	ice_candidate_t candidate;
-	if (ice_create_local_candidate(ICE_CANDIDATE_TYPE_HOST, 1, agent->local.candidates_count, record, &candidate, ICE_CANDIDATE_TRANSPORT_TCP_TYPE_ACTIVE)) {
-        JLOG_ERROR("Failed to create ice-tcp candidate");
-		return;
-	}
-
-	if (ice_add_candidate(&candidate, &agent->local)) {
-		JLOG_ERROR("Failed to add ice-tcp candidate to local description");
-	}
-
-	return;
-}
-
 int agent_gather_candidates(juice_agent_t *agent) {
 	JLOG_VERBOSE("Gathering candidates");
 	if (agent->conn_impl) {
@@ -283,10 +269,27 @@ int agent_gather_candidates(juice_agent_t *agent) {
 
 	if (agent->ice_tcp_mode == JUICE_ICE_TCP_MODE_ACTIVE) {
 		if (records_count > 0) {
-			addr_set_port((struct sockaddr *)&records[0].addr, 9);
-			agent_add_ice_tcp_local_candidate(agent, &records[0]);
+			// RFC 6544: The connection-address encoded into the candidate-attribute for
+			// active candidates MUST be set to the IP address that will be used for the
+			// attempt, but the port(s) MUST be set to 9 (i.e., Discard).
+			// Add only two TCP candidates, one for IPv6 and one for IPv4
+			const int families[2] = {AF_INET6, AF_INET};
+			for (int i = 0; i < 2; ++i) {
+				int family = families[i];
+				for (int j = 0; j < records_count; ++j) {
+					addr_record_t *record = records + j;
+					if (record->addr.ss_family == family) {
+						addr_record_t tcp_record;
+						memcpy(&tcp_record, record, sizeof(addr_record_t));
+						tcp_record.socktype = SOCK_STREAM;
+						addr_set_port((struct sockaddr *)&tcp_record.addr, 9);
+						agent_add_local_tcp_active_candidate(agent, &tcp_record);
+						break;
+					}
+				}
+			}
 		} else {
-			JLOG_WARN("No local host candidates gathered, unable to add ice-tcp");
+			JLOG_WARN("No local host candidates gathered, unable to add TCP candidates");
 		}
 	}
 
@@ -365,7 +368,8 @@ int agent_resolve_servers(juice_agent_t *agent) {
 			conn_unlock(agent);
 
 			addr_record_t records[DEFAULT_MAX_RECORDS_COUNT];
-			int records_count = addr_resolve(hostname, service, SOCK_DGRAM, records, DEFAULT_MAX_RECORDS_COUNT);
+			int records_count =
+			    addr_resolve(hostname, service, SOCK_DGRAM, records, DEFAULT_MAX_RECORDS_COUNT);
 
 			conn_lock(agent);
 
@@ -450,7 +454,8 @@ int agent_resolve_servers(juice_agent_t *agent) {
 		conn_unlock(agent);
 
 		addr_record_t records[MAX_STUN_SERVER_RECORDS_COUNT];
-		int records_count = addr_resolve(hostname, service, SOCK_DGRAM, records, MAX_STUN_SERVER_RECORDS_COUNT);
+		int records_count =
+		    addr_resolve(hostname, service, SOCK_DGRAM, records, MAX_STUN_SERVER_RECORDS_COUNT);
 
 		conn_lock(agent);
 
@@ -849,7 +854,8 @@ int agent_input(juice_agent_t *agent, char *buf, size_t len, const addr_record_t
 	return -1;
 }
 
-void agent_register_entry_for_candidate_pair(juice_agent_t *agent, ice_candidate_pair_t *pair, agent_stun_entry_t *relay_entry) {
+void agent_register_entry_for_candidate_pair(juice_agent_t *agent, ice_candidate_pair_t *pair,
+                                             agent_stun_entry_t *relay_entry) {
 	JLOG_VERBOSE("Registering STUN entry %d for candidate pair checking", agent->entries_count);
 	agent_stun_entry_t *entry = agent->entries + agent->entries_count;
 	entry->type = AGENT_STUN_ENTRY_TYPE_CHECK;
@@ -866,15 +872,32 @@ void agent_register_entry_for_candidate_pair(juice_agent_t *agent, ice_candidate
 		agent_translate_host_candidate_entry(agent, entry);
 }
 
-void agent_tcp_conn_connected(juice_agent_t *agent) {
+int agent_conn_tcp_state(juice_agent_t *agent, const addr_record_t *dst, tcp_state_t state) {
 	for (int i = 0; i < agent->entries_count; ++i) {
 		agent_stun_entry_t *entry = agent->entries + i;
-		if (entry->pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP) {
-			entry->pair->tcp_connected = true;
-			entry->next_transmission = current_timestamp();
+		if (entry->pair && entry->pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP &&
+		    addr_record_is_equal(&entry->record, dst, true)) {
+			entry->pair->tcp_state = state;
+			switch (state) {
+			case TCP_STATE_CONNECTED:
+				agent_arm_transmission(agent, entry, 0); // transmit now
+				break;
+			case TCP_STATE_DISCONNECTED:
+			case TCP_STATE_FAILED:
+				entry->pair->state = ICE_CANDIDATE_PAIR_STATE_FAILED;
+				entry->state = AGENT_STUN_ENTRY_STATE_FAILED;
+				entry->next_transmission = 0;
+				conn_interrupt(agent);
+				break;
+			default:
+				break;
+			}
+			return 0;
 		}
 	}
-	conn_interrupt(agent);
+
+	JLOG_WARN("No entry found for TCP address");
+	return -1;
 }
 
 int agent_bookkeeping(juice_agent_t *agent, timestamp_t *next_timestamp) {
@@ -889,20 +912,32 @@ int agent_bookkeeping(juice_agent_t *agent, timestamp_t *next_timestamp) {
 	for (int i = 0; i < agent->entries_count; ++i) {
 		agent_stun_entry_t *entry = agent->entries + i;
 
-		if (entry->pair && entry->pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP && entry->pair->tcp_connected == false) {
-			conn_tcp_connect(agent, &entry->pair->remote->resolved, agent_tcp_conn_connected);
-			entry->next_transmission = now + LAST_STUN_RETRANSMISSION_TIMEOUT;
-		} else if (entry->state == AGENT_STUN_ENTRY_STATE_PENDING) { // STUN requests transmission or retransmission
+		if (entry->state == AGENT_STUN_ENTRY_STATE_PENDING) { // STUN transmission or retransmission
 			if (entry->next_transmission > now)
 				continue;
 
+			if (entry->pair && entry->pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP) {
+			    if (entry->pair->tcp_state == TCP_STATE_DISCONNECTED)
+					conn_tcp_connect(agent, &entry->record); // First attempt a TCP connection
+
+				if(entry->pair->tcp_state != TCP_STATE_CONNECTED)
+					continue;
+			}
+
 			if (entry->retransmissions >= 0) {
 				if (JLOG_DEBUG_ENABLED) {
 					char record_str[ADDR_MAX_STRING_LEN];
 					addr_record_to_string(&entry->record, record_str, ADDR_MAX_STRING_LEN);
-					JLOG_DEBUG("STUN entry %d: Sending request to %s (%d retransmission%s left)", i,
-					           record_str, entry->retransmissions,
-					           entry->retransmissions >= 2 ? "s" : "");
+					if (entry->retransmissions == 0) {
+						JLOG_DEBUG("STUN entry %d: Sending request to %s (no retransmission)", i,
+						           record_str);
+					} else if (entry->retransmissions == 1) {
+						JLOG_DEBUG("STUN entry %d: Sending request to %s (1 retransmission left)",
+						           i, record_str);
+					} else {
+						JLOG_DEBUG("STUN entry %d: Sending request to %s (%d retransmissions left)",
+						           i, record_str, entry->retransmissions);
+					}
 				}
 				if (entry->transaction_id_expired) {
 					juice_random(entry->transaction_id, STUN_TRANSACTION_ID_SIZE);
@@ -1165,7 +1200,7 @@ int agent_bookkeeping(juice_agent_t *agent, timestamp_t *next_timestamp) {
 						if (entry->pair && entry->pair == selected_pair) {
 							entry->state =
 							    AGENT_STUN_ENTRY_STATE_PENDING;      // we don't want keepalives
-							entry->transaction_id_expired = true;	 // this is a new request
+							entry->transaction_id_expired = true;    // this is a new request
 							agent_arm_transmission(agent, entry, 0); // transmit now
 							break;
 						}
@@ -1356,7 +1391,8 @@ int agent_dispatch_stun(juice_agent_t *agent, void *buf, size_t size, stun_messa
 	case STUN_METHOD_ALLOCATE:
 	case STUN_METHOD_REFRESH:
 		if (agent_verify_credentials(agent, entry, buf, size, msg)) {
-			JLOG_WARN("Ignoring TURN Allocate message with invalid credentials (Is server authentication disabled?)");
+			JLOG_WARN("Ignoring TURN Allocate message with invalid credentials (Is server "
+			          "authentication disabled?)");
 			return -1;
 		}
 		return agent_process_turn_allocate(agent, msg, entry);
@@ -1950,7 +1986,7 @@ int agent_send_turn_allocate_request(juice_agent_t *agent, const agent_stun_entr
 	}
 
 	const char *password = NULL;
-	if(*entry->turn->credentials.nonce != '\0') {
+	if (*entry->turn->credentials.nonce != '\0') {
 		msg.credentials = entry->turn->credentials;
 		password = entry->turn->password;
 	}
@@ -2274,7 +2310,8 @@ int agent_add_local_reflexive_candidate(juice_agent_t *agent, ice_candidate_type
 		return 0;
 	}
 	ice_candidate_t candidate;
-	if (ice_create_local_candidate(type, 1, agent->local.candidates_count, record, &candidate, ICE_CANDIDATE_TRANSPORT_UDP)) {
+	if (ice_create_local_candidate(type, 1, agent->local.candidates_count, record, &candidate,
+	                               ICE_CANDIDATE_TRANSPORT_UDP)) {
 		JLOG_ERROR("Failed to create reflexive candidate");
 		return -1;
 	}
@@ -2314,7 +2351,8 @@ int agent_add_remote_reflexive_candidate(juice_agent_t *agent, ice_candidate_typ
 		return 0;
 	}
 	ice_candidate_t candidate;
-	if (ice_create_local_candidate(type, 1, agent->local.candidates_count, record, &candidate, ICE_CANDIDATE_TRANSPORT_UDP)) {
+	if (ice_create_local_candidate(type, 1, agent->local.candidates_count, record, &candidate,
+	                               ICE_CANDIDATE_TRANSPORT_UDP)) {
 		JLOG_ERROR("Failed to create reflexive candidate");
 		return -1;
 	}
@@ -2337,6 +2375,21 @@ int agent_add_remote_reflexive_candidate(juice_agent_t *agent, ice_candidate_typ
 	return agent_add_candidate_pairs_for_remote(agent, remote);
 }
 
+int agent_add_local_tcp_active_candidate(juice_agent_t *agent, addr_record_t *record) {
+	ice_candidate_t candidate;
+	if (ice_create_local_candidate(ICE_CANDIDATE_TYPE_HOST, 1, agent->local.candidates_count,
+	                               record, &candidate, ICE_CANDIDATE_TRANSPORT_TCP_TYPE_ACTIVE)) {
+		JLOG_ERROR("Failed to create TCP candidate");
+		return -1;
+	}
+	if (ice_add_candidate(&candidate, &agent->local)) {
+		JLOG_ERROR("Failed to add TCP candidate to local description");
+		return -1;
+	}
+
+	return 0;
+}
+
 int agent_add_candidate_pair(juice_agent_t *agent, ice_candidate_t *local, // local may be NULL
                              ice_candidate_t *remote) {
 	ice_candidate_pair_t pair;
@@ -2353,19 +2406,21 @@ int agent_add_candidate_pair(juice_agent_t *agent, ice_candidate_t *local, // lo
 
 	if (remote->transport != ICE_CANDIDATE_TRANSPORT_UDP) {
 		if (agent->ice_tcp_mode != JUICE_ICE_TCP_MODE_ACTIVE) {
-			JLOG_WARN("ICE-TCP is disabled ignoring TCP Candidate");
+			JLOG_WARN("ICE-TCP is disabled, ignoring TCP candidate");
 			return 0;
 		}
 
-		if (remote->transport != ICE_CANDIDATE_TRANSPORT_TCP_TYPE_PASSIVE && remote->transport != ICE_CANDIDATE_TRANSPORT_TCP_TYPE_SO) {
-			JLOG_INFO("Ignoring ICE-TCP Candidate that is not passive or simultaneous open");
+		if (remote->transport != ICE_CANDIDATE_TRANSPORT_TCP_TYPE_PASSIVE &&
+		    remote->transport != ICE_CANDIDATE_TRANSPORT_TCP_TYPE_SO) {
+			JLOG_INFO("TCP candidate is not passive or simultaneous open, ignoring");
 			return 0;
 		}
 
 		for (int i = 0; i < agent->candidate_pairs_count; ++i) {
 			ice_candidate_pair_t *pair = &agent->candidate_pairs[i];
-			if (pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP) {
-				JLOG_INFO("Only one ICE-TCP remote candidate is supported ignoring TCP Candidate");
+			if (pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP &&
+			    pair->remote->resolved.addr.ss_family == remote->resolved.addr.ss_family) {
+				JLOG_INFO("Only one remote TCP candidate is supported, ignoring");
 				return 0;
 			}
 		}
@@ -2501,10 +2556,15 @@ void agent_arm_transmission(juice_agent_t *agent, agent_stun_entry_t *entry, tim
 	entry->next_transmission = current_timestamp() + delay;
 
 	if (entry->state == AGENT_STUN_ENTRY_STATE_PENDING) {
-		entry->retransmission_timeout = MIN_STUN_RETRANSMISSION_TIMEOUT;
-		entry->retransmissions = entry->type == AGENT_STUN_ENTRY_TYPE_CHECK
-		                             ? MAX_STUN_CHECK_RETRANSMISSION_COUNT
-		                             : MAX_STUN_SERVER_RETRANSMISSION_COUNT;
+		if (entry->pair && entry->pair->remote->transport != ICE_CANDIDATE_TRANSPORT_UDP) {
+			entry->retransmission_timeout = STUN_TCP_TIMEOUT;
+			entry->retransmissions = 0; // do not retransmit
+		} else {
+			entry->retransmission_timeout = MIN_STUN_RETRANSMISSION_TIMEOUT;
+			entry->retransmissions = entry->type == AGENT_STUN_ENTRY_TYPE_CHECK
+			                             ? MAX_STUN_CHECK_RETRANSMISSION_COUNT
+			                             : MAX_STUN_SERVER_RETRANSMISSION_COUNT;
+		}
 	}
 
 	// Find a time slot
@@ -2684,8 +2744,7 @@ agent_stun_entry_t *agent_find_entry_from_record(juice_agent_t *agent, const add
 	return NULL;
 }
 
-int agent_set_ice_tcp_mode(juice_agent_t *agent, juice_ice_tcp_mode_t ice_tcp_mode)
-{
+int agent_set_ice_tcp_mode(juice_agent_t *agent, juice_ice_tcp_mode_t ice_tcp_mode) {
 	if (agent->conn_impl) {
 		JLOG_WARN("Unable to set ICE attributes, candidates gathering already started");
 		return JUICE_ERR_FAILED;
