@@ -183,6 +183,12 @@ int ice_create_local_candidate(ice_candidate_type_t type, int component, int ind
 	candidate->transport = transport;
 	strcpy(candidate->foundation, "-");
 
+	int socktype = transport == ICE_CANDIDATE_TRANSPORT_UDP ? SOCK_DGRAM : SOCK_STREAM;
+	if (record->socktype && record->socktype != socktype) {
+		JLOG_WARN("Incorrect socket type for candidate transport");
+	}
+	candidate->resolved.socktype = socktype;
+
 	candidate->priority = ice_compute_priority(candidate->type, candidate->resolved.addr.ss_family,
 	                                           candidate->component, index, candidate->transport);
 
@@ -205,7 +211,9 @@ int ice_resolve_candidate(ice_candidate_t *candidate, ice_resolve_mode_t mode) {
 		hints.ai_socktype = SOCK_DGRAM;
 		hints.ai_protocol = IPPROTO_UDP;
 	}
+#ifdef AI_ADDRCONFIG
 	hints.ai_flags = AI_ADDRCONFIG;
+#endif
 	if (mode != ICE_RESOLVE_MODE_LOOKUP)
 		hints.ai_flags |= AI_NUMERICHOST | AI_NUMERICSERV;
 	struct addrinfo *ai_list = NULL;
