@@ -11,12 +11,15 @@
 
 #include "addr.h"
 #include "juice.h"
+#include "tcp.h"
 #include "timestamp.h"
 
 #include <stdbool.h>
 #include <stdint.h>
 
-#define ICE_MAX_CANDIDATES_COUNT 20 // ~ 500B * 20 = 10KB
+#ifndef ICE_MAX_CANDIDATES_COUNT
+#define ICE_MAX_CANDIDATES_COUNT 30 // ~ 500B * 30 = 15KB
+#endif
 
 #define ICE_CANDIDATE_PENALTY_TCP 50
 
@@ -76,7 +79,7 @@ typedef struct ice_candidate_pair {
 	ice_candidate_pair_state_t state;
 	bool nominated;
 	bool nomination_requested;
-	bool tcp_connected;
+	tcp_state_t tcp_state;
 	timestamp_t consent_expiry;
 } ice_candidate_pair_t;
 
@@ -94,7 +97,8 @@ int ice_parse_sdp(const char *sdp, ice_description_t *description);
 int ice_parse_candidate_sdp(const char *line, ice_candidate_t *candidate);
 int ice_create_local_description(ice_description_t *description);
 int ice_create_local_candidate(ice_candidate_type_t type, int component, int index,
-                               const addr_record_t *record, ice_candidate_t *candidate, ice_candidate_transport_t transport);
+                               const addr_record_t *record, ice_candidate_t *candidate,
+                               ice_candidate_transport_t transport);
 int ice_resolve_candidate(ice_candidate_t *candidate, ice_resolve_mode_t mode);
 int ice_add_candidate(ice_candidate_t *candidate, ice_description_t *description);
 void ice_sort_candidates(ice_description_t *description);
@@ -109,7 +113,8 @@ int ice_update_candidate_pair(ice_candidate_pair_t *pair, bool is_controlling);
 
 int ice_candidates_count(const ice_description_t *description, ice_candidate_type_t type);
 
-uint32_t ice_compute_priority(ice_candidate_type_t type, int family, int component, int index, ice_candidate_transport_t transport);
+uint32_t ice_compute_priority(ice_candidate_type_t type, int family, int component, int index,
+                              ice_candidate_transport_t transport);
 
 bool ice_is_valid_string(const char *str);
 
